from typing import Optional, Sequence, Tuple

from _typeshed import Incomplete
from pydantic import BaseModel

logger: Incomplete

def camel_case(s: str) -> str: ...

class UrlModel(BaseModel):
    pattern_name: str
    namespace: str
    url_parts: Tuple[str, ...]
    js_vars: Tuple[str, ...]
    url_lookup: str
    is_alternative: bool
    def matches(self, name: str, params: Optional[Tuple[str]] = ...) -> bool: ...
    @property
    def js_func_name(self): ...
    @property
    def url(self): ...
    @property
    def js_literal(self): ...
    @property
    def as_function(self): ...
    @property
    def as_class_prop(self): ...
    @property
    def as_arrow_func(self): ...
    @property
    def as_arrow_func_property(self): ...

class UrlModels(BaseModel):
    urls: Sequence[UrlModel]
    def filtered_by_name(self, match=...): ...
    @classmethod
    def all_urls(cls): ...

def all_urls(): ...



import logging
from functools import lru_cache, reduce
from re import sub
from typing import Iterable, Optional, Sequence, Tuple, Union
from urllib.parse import urljoin

from pydantic import BaseModel

from django_lit_urls.utils import _parse_resolver

logger = logging.getLogger(__name__)


def camel_case(s: str) -> str:
    s = sub(r"(_|-)+", " ", s).title().replace(" ", "")
    return "".join([s[0].lower(), s[1:]])


class UrlModel(BaseModel):
    pattern_name: str
    namespace: str
    url_parts: Tuple[str, ...]
    js_vars: Tuple[str, ...]
    url_lookup: str
    is_alternative: bool = False  # Flag whether this is an "extra" pattern name,

    class Config:
        frozen = True

    def matches(self, name: str, params: Optional[Tuple[str]] = None) -> bool:
        """
        Check whether the name and optional parameter names passed
        """
        return self.pattern_name == name and self.js_vars == params if params else self.pattern_name == name

    @property
    def js_func_name(self):
        return camel_case(self.pattern_name)

    @property
    def url(self):
        """
        Return a string literal for
        the current URL incorporating the "js_vars" placeholders
        """
        return reduce(urljoin, self.url_parts)

    @property
    def js_literal(self):
        """
        Returns a JS string literal if there are variables, otherwise a string
        """
        return f"`{self.url}`" if self.js_vars else f'"{self.url}"'

    @property
    def as_function(self):
        """
        Return the template string
        as a standalone function
        """
        return f"function {self.js_func_name}({', '.join(self.js_vars)}){{ return {self.js_literal} }}"

    @property
    def as_class_prop(self):
        """
        Return the template string
        as a standalone function
        """
        return f"{self.js_func_name} ({', '.join(self.js_vars)}) {{ return {self.js_literal} }}"

    @property
    def as_arrow_func(self):
        return f"{self.js_func_name} = ({', '.join(self.js_vars)}) => {self.js_literal};"

    @property
    def as_arrow_func_property(self):
        return f"{self.js_func_name}: ({', '.join(self.js_vars)}) => {self.js_literal}"


class UrlModels(BaseModel):
    urls: Sequence[UrlModel]

    def filtered_by_name(self, matches: Iterable[Union[str, Tuple[str, Tuple[str]]]] = []):
        """
        Returns a new UrlModels instance
        where the URL names and optionally parameters match the
        given 'match' strings / tuples
        """
        return self.__class__(
            urls=set(
                [
                    um
                    for um in self.urls
                    for match in matches
                    if (isinstance(match, str) and um.matches(match))
                    or (not isinstance(match, str) and um.matches(*match))
                ]
            )
        )


@lru_cache()
def all_urls() -> UrlModels:
    return UrlModels(urls=[UrlModel(**_r) for _r in _parse_resolver()])



import logging
import re
import warnings
from typing import Any, Dict, Iterable, Optional, Set, Tuple, Union

from django.urls import LocalePrefixPattern, URLPattern, URLResolver, get_resolver
from django.urls.resolvers import RegexPattern, RoutePattern

logger = logging.getLogger(__name__)

MaybeResolver = Optional[Union[URLPattern, URLResolver]]
Url = str
JsFunction = str
Namespace = str
Params = Tuple[str, ...]

url_arg_re = re.compile(r"(\(.*?\))")
url_kwarg_re = re.compile(r"(\(\?P\<(.*?)\>.*?\))")
url_optional_char_re = re.compile(r"(?:\w|/)(?:\?|\*)")
url_optional_group_re = re.compile(r"\(\?\:.*\)(?:\?|\*)")
url_path_re = re.compile(r"(\<.*?\>)")


def _parse_resolver(
    resolver: Optional[URLResolver] = None,
    current_namespace: Optional[str] = None,
    url_parts: Tuple[str, ...] = ("/",),
    include_all: bool = True,
    include_namespaces: Set[str] = set(),
    exclude_namespaces: Set[str] = {
        "admin",
    },
    js_vars: Params = (),
    url_lookup: str = "",
) -> Iterable[Dict[str, Any]]:

    resolver = resolver or get_resolver()

    # Determine the effective namespace based on the
    # curent namespace and the resolver
    if current_namespace and resolver.namespace:
        ns = f"{current_namespace}:{resolver.namespace}"
    elif resolver.namespace:
        ns = f"{resolver.namespace}"
    elif current_namespace:
        ns = f"{current_namespace}"
    else:
        ns = ""

    if not (
        include_all
        and (include_namespaces is not None and ns not in include_namespaces)
        or (exclude_namespaces is not None and ns in exclude_namespaces)
    ):
        logger.debug(f"Skipping namespace {ns}")
        return

    for url_pattern in resolver.url_patterns:
        # url_part_or_full is a url with variables replaced by "Javascript" type literal vars
        # js_vars is the list of named variables in the function
        try:
            logger.debug(f"parsing {url_pattern}")
            standardise_url: str = _url_from_pattern(url_pattern)  # type: ignore
            url_part_or_full, extra_js_vars = _url_to_js_func(standardise_url)
            logger.debug(f"parsed {standardise_url}: {url_part_or_full} {extra_js_vars}")
        except Exception as E:
            warnings.warn(f"{E}")
            continue

        if isinstance(url_pattern, URLResolver):
            yield from _parse_resolver(
                url_pattern,
                current_namespace=ns,
                url_parts=(*url_parts, url_part_or_full),
                include_all=include_all,
                js_vars=(*js_vars, *extra_js_vars),
            )

        elif isinstance(url_pattern, URLPattern) and url_pattern.name:
            yield dict(
                pattern_name=url_pattern.name,
                namespace=ns,
                url_parts=(*url_parts, url_part_or_full),
                js_vars=(*js_vars, *extra_js_vars),
                url_lookup=f"{url_pattern.lookup_str}",
            )


def _url_from_pattern(url_pattern: Union[URLPattern, URLResolver]) -> str:
    """
    "Standardise" the different types of URL
    we might encounter to a simple string
    """

    pattern: Union[LocalePrefixPattern, RegexPattern, RoutePattern] = url_pattern.pattern
    url: Optional[str] = None

    if isinstance(pattern, LocalePrefixPattern):
        url = pattern.language_prefix
    elif isinstance(pattern, RegexPattern):
        url = pattern._regex  # type: ignore
    elif isinstance(pattern, RoutePattern):
        url = pattern._route  # type: ignore
    else:
        raise NotImplementedError(f"Unrecognized URL pattern type: {url_pattern} {type(pattern)}")
    if url is None:
        raise NotImplementedError(f"Unrecognized URL pattern type: {url_pattern} {type(pattern)}")
    return url.replace("^", "").replace("$", "")


def _url_to_js_func(url: Url) -> Tuple[JsFunction, Params]:
    js_vars = []
    new_url = f"{url}"
    for str_replace, js_var_name, js_var_pos in _url_js_replacements(url):
        logger.debug(f"{str_replace} -> {js_var_name} ({js_var_pos})")
        if js_var_name:
            new_url = new_url.replace(str_replace, f"${{{js_var_name}}}")
        else:
            # Assign an "anonymous" positional var
            js_var_name = f"v_{js_var_pos}"
            new_url = new_url.replace(str_replace, f"${{{js_var_name}}}")
        js_vars.append(js_var_name)
    return new_url, tuple(js_vars)


def _url_js_replacements(url: str) -> Iterable[Tuple[str, Optional[str], int]]:
    """
    Returns tuples of the strings to be replaced in a given URL
    """
    pos = 0
    logger.debug(url)
    # Removes optional groups from the URL pattern.
    if optional_group_matches := url_optional_group_re.findall(url):
        for el in optional_group_matches:
            pos += 1
            logger.debug(f"opt group matches {el} ({pos})")
            url = url.replace(el, "")
            yield el, None, pos

    # Removes optional characters from the URL pattern.
    if optional_char_matches := url_optional_char_re.findall(url):
        for el in optional_char_matches:
            pos += 1
            logger.debug(f"opt char matches {el} ({pos})")
            url = url.replace(el, "")
            yield el, None, pos

    # Identifies named URL arguments inside the URL pattern.
    # A "named URL argument" has the form `(?P<year>[0-9]{4})`
    if kwarg_matches := url_kwarg_re.findall(url):
        for el in kwarg_matches:
            pos += 1
            logger.debug(f"kwarg matches {el[0]} ({pos})")
            url = url.replace(el[0], "")
            yield el[0], el[1], pos

    # Identifies unnamed URL arguments inside the URL pattern.
    if args_matches := url_arg_re.findall(url):
        for el in args_matches:
            pos += 1
            logger.debug(f"arg matches {el} ({pos})")
            url = url.replace(el, "")
            yield el, None, pos

    # Identifies path expression and associated converters inside the URL pattern.
    # A "path  URL argument" has an optional converter type and parameter name.
    # Such as `<int:year>`
    if path_matches := url_path_re.findall(url):
        for el in path_matches:
            pos += 1
            replace_in_path = el.split(":")[-1].replace(":", "").replace("<", "").replace(">", "")
            logger.debug(f"path matches {el} ({pos})")
            yield el, replace_in_path, pos



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_lit_urls']

package_data = \
{'': ['*'], 'django_lit_urls': ['templates/*']}

install_requires = \
['Django>3.1', 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'django-lit-urls',
    'version': '0.1.2',
    'description': 'Django URLS delivered as string literal functions in Javascript',
    'long_description': None,
    'author': 'Josh Brooks',
    'author_email': 'josh@catalpa.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, Tuple, Union

from _typeshed import Incomplete as Incomplete
from django.urls import URLPattern, URLResolver

logger: Incomplete
MaybeResolver = Optional[Union[URLPattern, URLResolver]]
Url = str
JsFunction = str
Namespace = str
Params = Tuple[str, ...]
url_arg_re: Incomplete
url_kwarg_re: Incomplete
url_optional_char_re: Incomplete
url_optional_group_re: Incomplete
url_path_re: Incomplete
